==> sage-2005-03-03-0918-src-0.1.tar: sage-0.1/sage/rings/number_field.py <==
"""
Number field
"""

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import weakref

import arith
import ring
import sage.structure.gens as gens
import rational_field
import rational
import polynomial_ring
import number_field_element
import sage.pari.wrap

R = polynomial_ring.PolynomialRing(rational_field.RationalField())
x = R.gen()

_objsNumberField = {}
class _uniqNumberField(object):
    def __new__(cls, polynomial, variable="a", check=False):
        key = (cls, polynomial, variable)
        if _objsNumberField.has_key(key):
            x = _objsNumberField[key]()
            if x != None: return x
        if cls == CyclotomicField and polynomial <= 2:
            return rational_field.RationalField()
        elif cls == NumberField and polynomial.degree() <= 1:
            return rational_field.RationalField()
        R = object.__new__(cls)
        _objsNumberField[key] = weakref.ref(R)
        return R

class NumberField(_uniqNumberField, ring.Ring, gens.Gens):

    def __init__(self, polynomial, variable="a", check=False):
        if check and not polynomial.is_irreducible():
            raise ValueError, "defining polynomial must be irreducible"
        self.__polynomial = polynomial
        self.__degree = polynomial.degree()
        self.__variable = variable
        ring.Ring.__init__(self)

    def __repr__(self):
        return "Number Field with defining polynomial %s"%(
                   self.__polynomial)

    def __call__(self, x):
        """
        Coerce x into this number field.
        """
        if isinstance(x, number_field_element.NumberFieldElement):
            if x.parent() == self:
                return x
            raise TypeError,\
                  "Cannot coerce %s into %s"%(x,self)
        return number_field_element.NumberFieldElement(self, x)

    def __cmp__(self, other):
        if not isinstance(other, NumberField):
            return -1
        if self.__variable != other.__variable:
            return -1
        if self.__polynomial != other.__polynomial:
            return -1
        return 0

    def __contains__(self, x):
        return isinstance(x, number_field_element.NumberFieldElement) and \
               x.parent() == self

    def __get_pari_bnf(self):
        try:
            return self.__pari_bnf
        except AttributeError:
            self.__pari_bnf = sage.pari.wrap.PariObject("bnfinit(%s)"%self.polynomial())
            return self.__pari_bnf
    pari_bnf = property(__get_pari_bnf, \
            doc='PARI big number field corresponding to this field.')

    def characteristic(self):
        return 0

    def class_group(self):
        try:
            return self.__class_group
        except AttributeError:
            k = self.pari_bnf
            s = str(k[7][0])  # it's the [8][1] entry in pari, but the interface is 0 based.
            s = s.replace(";",",")
            s = eval(s)
            self.__class_group = sage.groups.abelian.AbelianGroup(s[1])
        return self.__class_group
            
    def degree(self):
        return self.__degree

    def disc(self, v):
        """
        The determinant of the trace pairing on the elements of the
        list v.
        """
        return self.trace_pairing(v).det()
        
    def gen(self, n=0):
        if n != 0:
            raise IndexError, "Only one generator."
        try:
            return self.__gen
        except:
            if self.__polynomial != None:
                X = self.__polynomial.parent().gen()
            else:
                X = PolynomialRing(rational_field.RationalField()).gen()
            self.__gen = number_field_element.NumberFieldElement(self, X)
            return self.__gen

    def is_field(self):
        return True

    def name(self):
        x="nf%s"%self.__polynomial
        x=x.replace(" ","")
        x=x.replace(operator.mul,"")
        return x

    def ngens(self):
        return 1
        
    def order_table(self):
        return []
        
    def polynomial(self):
        if self.__polynomial != None:
            return self.__polynomial
        if self.__is_cyclotomic:
            n = self.__zeta.order()
            self.__polynomial = PolynomialRing(rational_field.Q).cyclotomic_polynomial(n)
            return self.__polynomial
        raise NotImplementedError, "Don't know how to create defining polynomial."

    def polynomial_ring(self):
        return R
        
    def trace_pairing(self, v):
        """
        Returns the trace pairing on the elements of the list v.
        """
        A = linalg.MatrixSpace(RationalField(),len(v))(0)
        for i in range(len(v)):
            for j in range(i,len(v)):
                t = (v[i]*v[j]).trace()
                A[i,j] = t
                A[j,i] = t
        return A

    def fundamental_units(self):
        try:
            return self.__fundamental_units
        except AttributeError:
            k = self.pari_bnf
            s = k.bnfunit()
            t = str(s)
            t = t.replace("^","**")
            a = self.gen()
            t = t.replace("x", "a")
            s = eval(t)
            self.__fundamental_units = s
        return self.__fundamental_units
        
        
    def variable(self, new_var=None):
        """
        Returns or sets the name of the generator of the number field.

        If the optional string argument new_var is given, then the name of
        the number field generator is set to new_var.  If it is not given,
        this function returns the current name of the generator.
        
        INPUT:
            new_value -- optional str
        OUTPUT:
            str -- the name of the generator
        """
        if new_var==None:
            return self.__variable
        key = (self.__polynomial, self.__variable)
        if _objsNumberField.has_key(key):
            del _objsNumberField[key]
        self.__variable = new_var
        self.__repr = None
        _objsNumberField[(self.__polynomial, new_var)] = weakref.ref(self)
        
    def zeta(self):
        raise NotImplementedError


class CyclotomicField(NumberField):
    """
    Create a cyclotomic extension of the rational field.
    
    The command CyclotomicField(n) creates the n-th cyclotomic
    field, got by adjoing an n-th root of unity to the rational
    field.
    
    EXAMPLES:
        >>> CyclotomicField(3)
        Cyclotomic Field of order 3 and degree 2
        >>> CyclotomicField(18)
        Cyclotomic Field of order 18 and degree 6
        >>> z = CyclotomicField(6).gen(); z
        zeta_6
        >>> z**3
        -1
        >>> (1+z)**3
        6*zeta_6 - 3
    """
    def __init__(self, n):
        f = R.cyclotomic_polynomial(n)
        NumberField.__init__(self, f, variable="zeta_%s"%n, check=False)
        self.gen()._NumberFieldElement__order = n
        self.__order = n
        
    def __repr__(self):
        return "Cyclotomic Field of order %s and degree %s"%(
                self.order(), self.degree())
                
    def __call__(self, x):
        """
        EXAMPLES:
        The following example illustrates coercion from the cyclotomic
        field Q(zeta_42) to the cyclotomic field Q(zeta_6), in a case
        where such coercion is defined:
            >>> k42 = CyclotomicField(42)
            >>> k6 = CyclotomicField(6)
            >>> a = k42.g0
            >>> b = a**7
            >>> b
            zeta_42^7
            >>> k6(b)
            zeta_6
            >>> b**2
            zeta_42^7 - 1
            >>> k6(b**2)
            zeta_6 - 1
        """
        if isinstance(x, number_field_element.NumberFieldElement) and \
                isinstance(x.parent(), CyclotomicField):
            K = x.parent()
            n = K.order()
            m = self.order()
            if m % n == 0:   # easy case
                e = m/n
                f = x.polynomial()
                X = f.parent().gen()
                g = f(X**e)
            else:
                n = x.order()
                if m % n == 0:
                    # Harder case.  E.g., x = (zeta_42)^7 and
                    # self.__zeta = zeta_6, so it is possible to
                    # coerce x in, but not zeta_42 in.
                    # Algorithm:
                    #    1. Compute self.__zeta as an element
                    #       of K = parent of x.  Call this y.
                    #    2. Write x as a power r of y.
                    #       TODO: we do step two STUPIDLY.
                    #    3. Return self.__zeta to the power r.
                    y = K(self.zeta())
                    z = y
                    for r in xrange(y.order()):
                        if z == x:
                            return self.zeta()**(r+1)
                        z *= y
                raise TypeError, "Cannot coerce %s into %s"%(x,self)
            return number_field_element.NumberFieldElement(self, g)
        else:
            return NumberField.__call__(self, x)
                
    def order(self):
        return self.__order
        
    def order_table(self):
        try:
            return self.__order_table
        except AttributeError:
            t = {}
            x = self(1)
            n = self.order()
            m = 0
            zeta = self.zeta()
            # todo: this desperately needs to be optimized!!!
            for i in range(n):
                t[x.polynomial()] = n/arith.GCD(m,n)   # order of (zeta_n)**m
                x *= zeta
                m += 1
            self.__order_table = t
            return t
            
    def zeta(self):
        return self.gen()
    
class QuadraticField(NumberField):
    """
    Create a quadratic extension of the rational field.
    
    The command QuadraticExtension(a) creates the field Q(sqrt(a)).
    
    EXAMPLES:
    >>> QuadraticField(3)
    Quadratic Field with defining polynomial x^2 - 3
    >>> QuadraticField(-4)
    Quadratic Field with defining polynomial x^2 + 4
    """
    def __init__(self, a, variable="a", check=True):
        b = rational.Rational(a)
        f = x**2 - b
        NumberField.__init__(self, f, variable=variable, check=check)
        
    def __repr__(self):
        return NumberField.__repr__(self).replace("Number","Quadratic")
        
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])

==> sage-2005-03-16-1800-src-0.1.tar: sage-0.1/sage/rings/number_field.py <==
"""
Number field
"""

# TODO:
# Implement arithmetic in the field using a multiplication table,
# which is what most systems do.  This would speed things up
# massively.

#*****************************************************************************
#       Copyright (C) 2004 William Stein <was@math.harvard.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

import operator

import weakref

import arith
import ring
import sage.structure.gens as gens
import rational_field
import integer_ring
import rational
import polynomial_ring
import polynomial
import number_field_element
import sage.pari.wrap

Q = rational_field.RationalField()
Z = integer_ring.IntegerRing()
R = polynomial_ring.PolynomialRing(Q)
x = R.gen()

_objsNumberField = {}
class _uniqNumberField(object):
    def __new__(cls, poly, variable="a", check=False):
        key = (cls, poly, variable)
        if _objsNumberField.has_key(key):
            z = _objsNumberField[key]()
            if z != None: return z
        if cls == QuadraticField and isinstance(poly,int):
            poly = x**2 - poly
        #if cls == CyclotomicField and poly <= 2:
        #    return rational_field.RationalField()
        if cls == NumberField and poly.degree() <= 2:
            return QuadraticField(poly)
        #if cls == NumberField and poly.degree() <= 1:
        #    return rational_field.RationalField()
        R = object.__new__(cls)
        _objsNumberField[key] = weakref.ref(R)
        return R

class NumberField(_uniqNumberField, ring.Ring, gens.Gens):

    def __init__(self, poly, variable="a", check=False):
        if not isinstance(poly, polynomial.Polynomial):
            raise TypeError, "polynomial must be a Polynomial"
        if check and not poly.is_irreducible():
            raise ValueError, "defining polynomial (%s) must be irreducible"%poly
        self.__polynomial = poly
        self.__degree = poly.degree()
        self.__variable = variable
        ring.Ring.__init__(self)

    def __repr__(self):
        return "Number Field with defining polynomial %s"%(
                   self.__polynomial)

    def __call__(self, x):
        """
        Coerce x into this number field.
        """
        if isinstance(x, number_field_element.NumberFieldElement):
            if x.parent() == self:
                return x
            raise TypeError,\
                  "Cannot coerce %s into %s"%(x,self)
        return number_field_element.NumberFieldElement(self, x)

    def __cmp__(self, other):
        if not isinstance(other, NumberField):
            return -1
        if self.__variable != other.__variable:
            return -1
        if self.__polynomial != other.__polynomial:
            return -1
        return 0

    def __contains__(self, x):
        return isinstance(x, number_field_element.NumberFieldElement) and \
               x.parent() == self

    def __get_pari_nf(self):
        try:
            return self.__pari_nf
        except AttributeError:
            self.__pari_nf = sage.pari.wrap.PariObject("nfinit(%s)"%self.polynomial())
            return self.__pari_nf
    pari_nf = property(__get_pari_nf, \
            doc='PARI big number field corresponding to this field.')

    def __get_pari_bnf(self):
        try:
            return self.__pari_bnf
        except AttributeError:
            self.__pari_bnf = sage.pari.wrap.PariObject("bnfinit(%s)"%self.polynomial())
            return self.__pari_bnf
    pari_bnf = property(__get_pari_bnf, \
            doc='PARI big number field corresponding to this field.')

    def characteristic(self):
        return 0

    def class_group(self):
        try:
            return self.__class_group
        except AttributeError:
            k = self.pari_bnf
            s = str(k[7][0])  # it's the [8][1] entry in pari, but the interface is 0 based.
            s = s.replace(";",",")
            s = eval(s)
            self.__class_group = sage.groups.abelian.AbelianGroup(s[1])
        return self.__class_group

    def class_number(self):
        return self.class_group().order()
            
    def degree(self):
        return self.__degree

    def disc(self, v=None):
        """
        Returns the discriminant of the ring of integers of the number field,
        or if v is specified, the determinant of the trace pairing
        on the elements of the list v.

        INPUT:
            v (optional) -- list of element of this number field
        OUTPUT:
            Integer if v is omitted, and Rational otherwise.
        """
        if v == None:
            try:
                return self.__disc
            except AttributeError:
                v = self.integral_basis()
                #k = self.__pari_bnf
                #self.__disc = Z(long(k[6][2]))
                self.__disc = Q(self.trace_pairing(v).det())
                return self.__disc
        else:
            return Q(self.trace_pairing(v).det())

    def discriminant(self, v=None):
        """
        Same as self.disc.
        """
        return self.disc(v)

    def factor_integer(self, n):
        """
        Ideal factorization of the principal ideal of the ring
        of integers generated by n.
        """
        F = self.pari_nf.idealfactor(n).list()
        return [FractionalIdeal(self, X) for X in F]

    def gen(self, n=0):
        if n != 0:
            raise IndexError, "Only one generator."
        try:
            return self.__gen
        except:
            if self.__polynomial != None:
                X = self.__polynomial.parent().gen()
            else:
                X = PolynomialRing(rational_field.RationalField()).gen()
            self.__gen = number_field_element.NumberFieldElement(self, X)
            return self.__gen

    def is_field(self):
        return True

    def integral_basis(self):
        """
        Return a list of elements of this number field that are a basis
        for the full ring of integers.

        
        TODO: Serious problem / bug -- this is computed using PARI,
              and the evals to turn the polys into elements of the
              number field completely fails when there are denominators!
        """
        try:
            return self.__integral_basis
        except AttributeError:
            #k = self.pari_bnf
            #B = k[6][6]
            B = sage.pari.pari_pyx.eval("nfbasis(%s)"%self.polynomial())
            if B.find("/") != -1:
                raise RuntimeError, "Computation of integral basis not done when there are denominators."
            x = self.gen()
            B = eval(str(B).replace("^","**"))
            B = [self(z) for z in B]
            self.__integral_basis = B
        return self.__integral_basis
    
    def name(self):
        x="nf%s"%self.__polynomial
        x=x.replace(" ","")
        x=x.replace(operator.mul,"")
        return x

    def ngens(self):
        return 1
        
    def order_table(self):
        return []
        
    def polynomial(self):
        if self.__polynomial != None:
            return self.__polynomial
        if self.__is_cyclotomic:
            n = self.__zeta.order()
            self.__polynomial = PolynomialRing(rational_field.Q).cyclotomic_polynomial(n)
            return self.__polynomial
        raise NotImplementedError, "Don't know how to create defining polynomial."

    def polynomial_ring(self):
        return R
        
    def trace_pairing(self, v):
        """
        Returns the trace pairing on the elements of the list v.
        """
        import sage.linalg.linalg
        A = sage.linalg.linalg.MatrixSpace(Q,len(v))(0)
        for i in range(len(v)):
            for j in range(i,len(v)):
                t = (v[i]*v[j]).trace()
                A[i,j] = t
                A[j,i] = t
        return A

    def fundamental_units(self):
        try:
            return self.__fundamental_units
        except AttributeError:
            k = self.pari_bnf
            s = k.bnfunit()
            t = str(s)
            t = t.replace("^","**")
            a = self.gen()
            t = t.replace("x", "a")
            s = eval(t)
            self.__fundamental_units = s
        return self.__fundamental_units
        
        
    def variable(self, new_var=None):
        """
        Returns or sets the name of the generator of the number field.

        If the optional string argument new_var is given, then the name of
        the number field generator is set to new_var.  If it is not given,
        this function returns the current name of the generator.
        
        INPUT:
            new_value -- optional str
        OUTPUT:
            str -- the name of the generator
        """
        if new_var==None:
            return self.__variable
        key = (self.__polynomial, self.__variable)
        if _objsNumberField.has_key(key):
            del _objsNumberField[key]
        self.__variable = new_var
        self.__repr = None
        _objsNumberField[(self.__polynomial, new_var)] = weakref.ref(self)
        
    def zeta(self):
        raise NotImplementedError


class CyclotomicField(NumberField):
    """
    Create a cyclotomic extension of the rational field.
    
    The command CyclotomicField(n) creates the n-th cyclotomic
    field, got by adjoing an n-th root of unity to the rational
    field.
    
    EXAMPLES:
        >>> CyclotomicField(3)
        Cyclotomic Field of order 3 and degree 2
        >>> CyclotomicField(18)
        Cyclotomic Field of order 18 and degree 6
        >>> z = CyclotomicField(6).gen(); z
        zeta_6
        >>> z**3
        -1
        >>> (1+z)**3
        6*zeta_6 - 3
    """
    def __init__(self, n):
        f = R.cyclotomic_polynomial(n)
        NumberField.__init__(self, f, variable="zeta_%s"%n, check=False)
        self.gen()._NumberFieldElement__order = n
        self.__order = n
        
    def __repr__(self):
        return "Cyclotomic Field of order %s and degree %s"%(
                self.order(), self.degree())
                
    def __call__(self, x):
        """
        EXAMPLES:
        The following example illustrates coercion from the cyclotomic
        field Q(zeta_42) to the cyclotomic field Q(zeta_6), in a case
        where such coercion is defined:
            >>> k42 = CyclotomicField(42)
            >>> k6 = CyclotomicField(6)
            >>> a = k42.g0
            >>> b = a**7
            >>> b
            zeta_42^7
            >>> k6(b)
            zeta_6
            >>> b**2
            zeta_42^7 - 1
            >>> k6(b**2)
            zeta_6 - 1
        """
        if isinstance(x, number_field_element.NumberFieldElement) and \
                isinstance(x.parent(), CyclotomicField):
            K = x.parent()
            n = K.order()
            m = self.order()
            if m % n == 0:   # easy case
                e = m/n
                f = x.polynomial()
                X = f.parent().gen()
                g = f(X**e)
            else:
                n = x.order()
                if m % n == 0:
                    # Harder case.  E.g., x = (zeta_42)^7 and
                    # self.__zeta = zeta_6, so it is possible to
                    # coerce x in, but not zeta_42 in.
                    # Algorithm:
                    #    1. Compute self.__zeta as an element
                    #       of K = parent of x.  Call this y.
                    #    2. Write x as a power r of y.
                    #       TODO: we do step two STUPIDLY.
                    #    3. Return self.__zeta to the power r.
                    y = K(self.zeta())
                    z = y
                    for r in xrange(y.order()):
                        if z == x:
                            return self.zeta()**(r+1)
                        z *= y
                raise TypeError, "Cannot coerce %s into %s"%(x,self)
            return number_field_element.NumberFieldElement(self, g)
        else:
            return NumberField.__call__(self, x)

    def integral_basis(self):
        """
        Return a list of elements of this number field that are a basis
        for the full ring of integers.
        """
        try:
            return self.__integral_basis
        except AttributeError:
            z = self.gen()
            a = self(1)
            B = []
            for n in xrange(self.degree()):
                B.append(a)
                a *= z
            self.__integral_basis = B
        return self.__integral_basis
        
                
    def order(self):
        return self.__order
        
    def order_table(self):
        try:
            return self.__order_table
        except AttributeError:
            t = {}
            x = self(1)
            n = self.order()
            m = 0
            zeta = self.zeta()
            # todo: this desperately needs to be optimized!!!
            for i in range(n):
                t[x.polynomial()] = n/arith.GCD(m,n)   # order of (zeta_n)**m
                x *= zeta
                m += 1
            self.__order_table = t
            return t
            
    def zeta(self):
        return self.gen()
    
class QuadraticField(NumberField):
    """
    Create a quadratic extension of the rational field.
    
    The command QuadraticExtension(a) creates the field Q(sqrt(a)).
    
    EXAMPLES:
    >>> QuadraticField(3)
    Quadratic Field with defining polynomial x^2 - 3
    >>> QuadraticField(-4)
    Quadratic Field with defining polynomial x^2 + 4
    """
    def __init__(self, f, variable="a", check=True):
        if not (f in R):
            f = x**2 - R(f)
        NumberField.__init__(self, f, variable=variable, check=check)
        self.__disc = eval(sage.pari.pari_pyx.eval("nfdisc(%s)"%f))
        
    def __repr__(self):
        return NumberField.__repr__(self).replace("Number","Quadratic")

    def class_number(self, proof = True):
        """
        Return the size of the class group of self.

        If proof = False (not the default) and the discriminant of the
        field is negative, then the following warning from the PARI
        manual applies: IMPORTANT WARNING: For D<0, this function may
        give incorrect results when the class group has a low exponent
        (has many cyclic factors), because implementing Shank's method
        in full generality slows it down immensely.
        """
        try:
            return self.__class_number
        except:
            D = self.discriminant()
            if D < 0 and proof:
                self.__class_number = eval(sage.pari.pari_pyx.eval("qfbclassno(%s,1)"%D))
            else:
                self.__class_number = eval(sage.pari.pari_pyx.eval("qfbclassno(%s)"%D))
            return self.__class_number

    def disc(self, v=None):
        if v==None:
            return self.__disc
        return NumberField.disc(v)

def is_fundamental_discriminant(D):
    d = D % 4
    if not (d in [0,1]):
        return False
    return D != 1 and  D != 0 and \
           (arith.is_squarefree(D) or \
            (d == 0 and (D/4)%4 in [2,3] and arith.is_squarefree(D/4)))

class FractionalIdeal:
    def __init__(self, number_field, defining_data):
        self.__defining_data = defining_data
        self.__number_field = number_field

    def __repr__(self):
        return "Fractional ideal defined by %s"%(self.__defining_data)

    def number_field(self):
        return self.__nf



        
if __name__ ==  '__main__':
    import doctest, sys
    doctest.testmod(sys.modules[__name__])
